/**************************************************************************
 * 
 * PROJECT: Task Control Management (TCM)
 *
 * FILE: CSFsmResponder.cc
 *
 * ABSTRACT: Object for responding using CSFsmResponseModuleClass.
 *
 * EXPORTS:
 *
 * $Revision: 1.2 $
 * $Date: 1998/03/18 14:41:45 $
 * $Author: reids $
 *
 * REVISION HISTORY:
 *
 * $Log: CSFsmResponder.cc,v $
 * Revision 1.2  1998/03/18 14:41:45  reids
 * Changes to support both SunOS and Solaris versions.
 *
// Revision 1.1  97/11/21  14:04:57  reids
// First release of TCM -- seems to be a stable version
// 
 *
 **************************************************************************/

#include <rtilib/makeheader.h>
#include <cs.h>
#include <CSFsm.h>

#include "CSFsmResponseModule.pkg.h"
#include "CSFsmResponder.h"

static CSFsmResponseModule _noStimResponse = NULL;

// Add thi to the list of responders 

void CSFsmResponderClass::add (CSFsm fsm)
{
  CSFsmResponder first = (CSFsmResponder)fsm->UserData();
  if (first) _next = first;
  fsm->UserData(this);

  // Set up mechanism to catch unhandled stimulus
  if (_stimulus && *_stimulus != '\0')
    setMissingStimulusResponse(fsm, _stimulus);
}

/* This is called to enable the fsm to respond correctly in case a stimulus
   is sent that cannot be handled in the current state.*/
void CSFsmResponderClass::setMissingStimulusResponse (CSFsm fsm,
						      String stimulus)
{
  if (!_noStimResponse) {
    _noStimResponse = new CSFsmResponseModuleClass(fsm, NO_STIM_MODULE);
    _noStimResponse->setContext(NO_STIM_MODULE);
  }
  fsm->TransitionInstall(CSFsmClass::STATE, CSFSM_ANY_STATE,
			 stimulus, _noStimResponse, CSFSM_END);
}

