
/** This class establishes a one-to-one mapping between names and objects.
 *  Objects may not be registered under more than one name.  No two objects
 *  may be registered under the same name.
 */

import java.util.Enumeration;

public class Registry extends Object
{
  protected DataHashtable  nameToObjectHashtable;
  protected DataHashtable  objectToNameHashtable;


  public Registry ( )
  {
    nameToObjectHashtable = new DataHashtable();
    objectToNameHashtable = new DataHashtable();
  }


  protected DataHashtable getNameToObjectHashtable ( )
					{ return nameToObjectHashtable; }
  protected DataHashtable getObjectToNameHashtable ( )
					{ return objectToNameHashtable; }


  public boolean register ( String theName, Object theObject )
  {
	/* Is this name/object pair already registered together? */
    if (   ( getObjectForName ( theName   ) == theObject )
	&& ( getNameForObject ( theObject ) == theName   ) )
    {
      System.err.println ( "[Registry:register]  Warning:  Object/Name already"
			   + " registered together.  (\"" + theName + "\")" );
      return true;  /* Already done before we start... */
    }

	/* If this name is already in use... */
    if ( getIsNameRegistered ( theName ) )
    {
      System.err.println ( "[Registry:register]  Warning:  Name already "
			   + "registered.  (\"" + theName + "\")" );
      return false;
    }

	/* If this object is already in use... */
    if ( getIsObjectRegistered ( theObject ) )
    {
      System.err.println ( "[Registry:register]  Warning:  Object already "
			   + "registered under name \""
			   + getNameForObject ( theObject ) + "\"." );
      return false;
    }

    getNameToObjectHashtable() . put ( theName, theObject );
    getObjectToNameHashtable() . put ( theObject, theName );
    return true;
  }


  public boolean unregister ( String theName )
  {
	/* If this String is not already in use */
    if ( getIsNameRegistered ( theName ) == false )
    {
      System.err.println ( "[Registry:unregister]  Warning:  Name is not "
			   + "registered.  (\"" + theName + "\")" );
      return false;
    }

    Object  theObject = getObjectForName ( theName );

    getNameToObjectHashtable() . remove ( theName   );
    getObjectToNameHashtable() . remove ( theObject );
    return true;
  }


  public boolean unregister ( Object theObject )
  {
	/* If this Object is not already in use */
    if ( getIsObjectRegistered ( theObject ) == false )
    {
      System.err.println ( "[Registry:unregister]  Warning:  "
			   + "Object is not registered." );
      return false;
    }

    String  theName = getNameForObject ( theObject );

    getNameToObjectHashtable() . remove ( theName   );
    getObjectToNameHashtable() . remove ( theObject );
    return true;
  }


  public Object  getObjectForName ( String theName )
  {
    return getNameToObjectHashtable() . get ( theName );
  }


  public String  getNameForObject ( Object theObject )
  {
    return (String) getObjectToNameHashtable() . get ( theObject );
  }


  public boolean getIsNameRegistered ( String theName )
  {
    return getObjectForName ( theName ) != null;
  }

  public boolean getIsObjectRegistered ( Object theObject )
  {
    return getNameForObject ( theObject ) != null;
  }

  public Enumeration getObjects()
  {
    return getObjectToNameHashtable() . keys();
  }

  public Enumeration getNames()
  {
    return getNameToObjectHashtable() . keys();
  }

  public String toString()
  {
    StringBuffer  stringBuffer = new StringBuffer();
    Enumeration   names = getNames();

    stringBuffer . append ( "Registry " );
    stringBuffer . append ( Integer.toHexString ( hashCode() ) );
    stringBuffer . append ( ":\n" );
    while ( names . hasMoreElements() )
    {
      stringBuffer . append ( "     \"" );
      stringBuffer . append ( (String) (names.nextElement()) );
      stringBuffer . append ( "\"\n" );
    }
    stringBuffer . append ( "\n" );
    return stringBuffer . toString();
  }
}
  
